from typing import List

from pydantic import BaseModel
from tortoise.contrib.pydantic import pydantic_model_creator, pydantic_queryset_creator

from rearq.server.models import Job, JobResult


class JobListOut(BaseModel):
    rows: pydantic_queryset_creator(Job)
    total: int


JobOut = pydantic_model_creator(Job)


class JobResultOut(pydantic_model_creator(JobResult)):
    job: JobOut


class JobResultListOut(BaseModel):
    rows: List[JobResultOut]
    total: int



import datetime
from typing import Any, Callable, Dict, Optional, Tuple, Union
from uuid import uuid4

from crontab import CronTab
from loguru import logger
from tortoise import timezone

from rearq.constants import DELAY_QUEUE
from rearq.job import JobStatus
from rearq.server.models import Job, JobResult
from rearq.utils import ms_to_datetime, timestamp_ms_now, to_ms_timestamp


class Task:
    def __init__(
        self,
        bind: bool,
        function: Callable,
        queue: str,
        rearq,
        job_retry: int,
        job_retry_after: int,
        expire: Optional[Union[float, datetime.datetime]] = None,
    ):

        self.job_retry = job_retry
        self.job_retry_after = job_retry_after
        self.queue = queue
        self.rearq = rearq
        self.function = function
        self.bind = bind
        self.expire = expire

    async def delay(
        self,
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
        job_id: str = None,
        countdown: Union[float, datetime.timedelta] = 0,
        eta: Optional[datetime.datetime] = None,
        expire: Optional[Union[float, datetime.datetime]] = None,
        job_retry: int = 0,
        job_retry_after: int = 60,
    ) -> Job:
        if not job_id:
            job_id = uuid4().hex
        if countdown:
            defer_ts = to_ms_timestamp(countdown)
        elif eta:
            defer_ts = to_ms_timestamp(eta)
        else:
            defer_ts = timestamp_ms_now()
        expire_time = None
        expires = expire or self.expire
        if expires:
            expire_time = ms_to_datetime(to_ms_timestamp(expires))

        job = await Job.get_or_none(job_id=job_id)
        if job:
            logger.warning(f"Job {job_id} exists")
            return job

        job = Job(
            task=self.function.__name__,
            args=args,
            kwargs=kwargs,
            job_retry=job_retry or self.job_retry,
            queue=self.queue,
            job_id=job_id,
            expire_time=expire_time,
            enqueue_time=timezone.now(),
            job_retry_after=job_retry_after,
        )

        if not eta and not countdown:
            job.status = JobStatus.queued
            await job.save()
            await self.rearq.redis.xadd(self.queue, {"job_id": job_id})
        else:
            job.status = JobStatus.deferred
            await job.save()
            await self.rearq.redis.zadd(DELAY_QUEUE, defer_ts, f"{self.queue}:{job_id}")

        return job


async def check_pending_msgs(self: Task, queue: str, group_name: str, timeout: int):
    """
    check pending messages
    :return:
    """
    redis = self.rearq.redis
    pending_msgs = await redis.xpending(self.queue, group_name, "-", "+", 10)
    p = redis.pipeline()
    execute = False
    for msg in pending_msgs:
        msg_id, _, idle_time, times = msg
        job_result = await JobResult.filter(msg_id=msg_id).only("job_id").first()
        if not job_result:
            continue
        if int(idle_time / 1000) > timeout * 2:
            execute = True
            p.xack(queue, group_name, msg_id)
            p.xadd(queue, {"job_id": job_result.job_id})
    if execute:
        return await p.execute()


class CronTask(Task):
    _cron_tasks: Dict[str, "CronTask"] = {}
    next_run: int

    def __init__(
        self,
        bind: bool,
        function: Callable,
        queue: str,
        rearq,
        job_retry: int,
        job_retry_after: int,
        cron: str,
        expire: Optional[Union[float, datetime.datetime]] = None,
        run_at_start: Optional[bool] = False,
    ):
        super().__init__(bind, function, queue, rearq, job_retry, job_retry_after, expire)
        self.crontab = CronTab(cron)
        self.cron = cron
        self.run_at_start = run_at_start
        self.set_next()

    def set_next(self):
        self.next_run = to_ms_timestamp(self.crontab.next(default_utc=False))

    @classmethod
    def add_cron_task(cls, function: str, cron_task: "CronTask"):
        cls._cron_tasks[function] = cron_task

    @classmethod
    def get_cron_tasks(cls):
        return cls._cron_tasks



from enum import Enum


class JobStatus(str, Enum):
    """
    Enum of job statuses.
    """

    deferred = "deferred"
    queued = "queued"
    in_progress = "in_progress"
    success = "success"
    failed = "failed"
    expired = "expired"



import asyncio
import importlib
import sys
from functools import wraps

import click
import uvicorn
from click import BadArgumentUsage, Context

from rearq.server.app import app
from rearq.version import VERSION
from rearq.worker import TimerWorker, Worker


def coro(f):
    @wraps(f)
    def wrapper(*args, **kwargs):
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(f(*args, **kwargs))

    return wrapper


@click.group(context_settings={"help_option_names": ["-h", "--help"]})
@click.version_option(VERSION, "-v", "--version")
@click.option("--verbose", default=False, is_flag=True, help="Enable verbose output.")
@click.argument("rearq", required=True)
@click.pass_context
@coro
async def cli(ctx: Context, rearq: str, verbose):
    splits = rearq.split(":")
    rearq_path = splits[0]
    rearq = splits[1]
    try:
        module = importlib.import_module(rearq_path)
        rearq = getattr(module, rearq, None)

        ctx.ensure_object(dict)
        ctx.obj["rearq"] = rearq
        ctx.obj["verbose"] = verbose

    except (ModuleNotFoundError, AttributeError) as e:
        raise BadArgumentUsage(ctx=ctx, message=f"Init rearq error, {e}.")


@cli.command(help="Start a worker.")
@click.option("-q", "--queue", required=False, help="Queue to consume.")
@click.option(
    "--group-name", required=False, help="Group name.",
)
@click.option("--consumer-name", required=False, help="Consumer name.")
@click.pass_context
@coro
async def worker(ctx: Context, queue: str, group_name: str, consumer_name: str):
    rearq = ctx.obj["rearq"]
    await rearq.init()
    w = Worker(rearq, queue=queue, group_name=group_name, consumer_name=consumer_name)
    await w.run()


@cli.command(help="Start a timer.")
@click.pass_context
@coro
async def timer(ctx: Context):
    rearq = ctx.obj["rearq"]
    await rearq.init()
    w = TimerWorker(rearq)
    await w.run()


@cli.command(help="Start rest api server.")
@click.option("--host", default="127.0.0.1", show_default=True, help="Listen host.")
@click.option("-p", "--port", default=8000, show_default=True, help="Listen port.")
@click.pass_context
def server(ctx: Context, host: str, port: int):
    rearq = ctx.obj["rearq"]
    app.rearq = rearq

    verbose = ctx.obj["verbose"]

    @app.on_event("startup")
    async def startup():
        await rearq.init()

    @app.on_event("shutdown")
    async def shutdown():
        await rearq.close()

    uvicorn.run(app=app, host=host, port=port, debug=verbose)


def main():
    sys.path.insert(0, ".")
    cli()


if __name__ == "__main__":
    main()



from starlette.requests import Request

from rearq import ReArq


async def get_rearq(request: Request) -> ReArq:
    return request.app.rearq


def get_pager(limit: int = 10, offset: int = 0):
    return limit, offset



from fastapi import APIRouter

from .index import router as index_router
from .job import router as job_router
from .result import router as result_router
from .task import router as task_router
from .worker import router as worker_router

router = APIRouter()

router.include_router(index_router, tags=["Index"])
router.include_router(job_router, prefix="/job", tags=["Job"])
router.include_router(task_router, prefix="/task", tags=["Task"])
router.include_router(worker_router, prefix="/worker", tags=["Worker"])
router.include_router(result_router, prefix="/result", tags=["Result"])



import asyncio
import json
import signal
import socket
from functools import partial
from signal import Signals
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Set
from uuid import uuid4

import async_timeout
from aioredis.errors import BusyGroupError
from aioredlock import Aioredlock
from loguru import logger
from tortoise import timezone
from tortoise.expressions import F

from rearq import CronTask, UsageError, constants
from rearq.constants import DEFAULT_QUEUE, DELAY_QUEUE, QUEUE_KEY_PREFIX
from rearq.job import JobStatus
from rearq.server.models import Job, JobResult
from rearq.task import check_pending_msgs
from rearq.utils import args_to_string, ms_to_datetime, poll, timestamp_ms_now, to_ms_timestamp

if TYPE_CHECKING:
    from rearq import ReArq
Serializer = Callable[[Dict[str, Any]], bytes]
Deserializer = Callable[[bytes], Dict[str, Any]]


class Worker:
    _task_map = {}

    def __init__(
        self,
        rearq: "ReArq",
        queue: Optional[str] = None,
        group_name: Optional[str] = None,
        consumer_name: Optional[str] = None,
    ):
        self.group_name = group_name or socket.gethostname()
        self.consumer_name = consumer_name
        self.job_timeout = rearq.job_timeout
        self.max_jobs = rearq.max_jobs
        self.rearq = rearq
        self._redis = rearq.redis
        self._lock_manager = Aioredlock([(rearq.redis_host, rearq.redis_port)])
        self.register_tasks = rearq.get_queue_tasks(queue)
        if queue:
            self.queue = QUEUE_KEY_PREFIX + queue
        else:
            self.queue = DEFAULT_QUEUE
        self.loop = asyncio.get_event_loop()
        self.sem = asyncio.BoundedSemaphore(self.max_jobs)
        self.queue_read_limit = max(self.max_jobs * 5, 100)
        self.tasks: Set[asyncio.Task[Any]] = set()
        self._task_map = rearq.task_map
        self.jobs_complete = 0
        self.jobs_retried = 0
        self.jobs_failed = 0
        self.job_retry = rearq.job_retry
        self.job_retry_after = rearq.job_retry_after
        self._add_signal_handler(signal.SIGINT, self.handle_sig)
        self._add_signal_handler(signal.SIGTERM, self.handle_sig)
        self.rearq.create_task(True, check_pending_msgs, queue, "* * * * *")

    def _add_signal_handler(self, signum: Signals, handler: Callable[[Signals], None]) -> None:
        self.loop.add_signal_handler(signum, partial(handler, signum))

    def handle_sig(self, signum: Signals) -> None:
        sig = Signals(signum)
        logger.info(
            "shutdown on %s ◆ %d jobs complete ◆ %d failed ◆ %d retries ◆ %d ongoing to cancel"
            % (sig.name, self.jobs_complete, self.jobs_failed, self.jobs_retried, len(self.tasks),)
        )
        for t in asyncio.all_tasks():
            t.cancel()

    async def log_redis_info(self,) -> None:
        p = self._redis.pipeline()
        p.info()
        p.dbsize()
        info, key_count = await p.execute()

        logger.info(
            f'redis_version={info["server"]["redis_version"]} '
            f'mem_usage={info["memory"]["used_memory_human"]} '
            f'clients_connected={info["clients"]["connected_clients"]} '
            f"db_keys={key_count}"
        )

    async def _main(self) -> None:
        logger.add(f"logs/worker-{self.worker_name}.log", rotation="00:00")
        logger.info(f"Start worker success with queue: {self.queue}")
        logger.info(f"Registered tasks: {', '.join(self.register_tasks)}")
        await self.log_redis_info()
        await self.rearq.startup()
        with await self._redis as redis:
            while True:
                msgs = await redis.xread_group(
                    self.group_name,
                    self.consumer_name,
                    [self.queue],
                    count=self.queue_read_limit,
                    latest_ids=[">"],
                )
                async with self.sem:
                    jobs_id = list(map(lambda m: m[2].get("job_id"), msgs))
                    qs = Job.filter(job_id__in=jobs_id)
                    await qs.update(status=JobStatus.in_progress)
                    jobs = await qs
                    jobs_dict = {job.job_id: job for job in jobs}
                    for msg in msgs:
                        await self.sem.acquire()
                        queue, msg_id, job = msg
                        job_id = job.get("job_id")
                        job = jobs_dict.get(job_id)
                        if not job:
                            logger.warning(f"job {job_id} not found")
                            await self._xack(queue, msg_id)
                            self.sem.release()
                            continue
                        task = self.loop.create_task(self.run_job(queue, msg_id, job))
                        task.add_done_callback(self._task_done)
                        self.tasks.add(task)
                await asyncio.gather(*self.tasks)

    def _task_done(self, task):
        self.sem.release()
        self.tasks.remove(task)

    async def _xack(self, queue: str, msg_id: str):
        await self._redis.xack(queue, self.group_name, msg_id)

    async def run_job(self, queue: str, msg_id: str, job: Job):
        if job.expire_time and job.expire_time > timezone.now():
            logger.warning(f"job {job.job_id} is expired, ignore")
            job.status = JobStatus.expired
            await job.save(update_fields=["status"])
            return
        redis = self._redis
        job_id = job.job_id
        job_result = JobResult(
            msg_id=msg_id, job=job, worker=self.worker_name, start_time=timezone.now()
        )
        task = self._task_map.get(job.task)
        if not task:
            logger.warning(f"job {job_id}, task {job.task} not found")
            job_result.result = "task not found"
            await job_result.save()
            return job_result
        ref = f"{job_id}:{job.task}"

        start_ms = timestamp_ms_now()
        logger.info(
            "%6.2fs → %s(%s)%s"
            % (
                (start_ms - to_ms_timestamp(job.enqueue_time)) / 1000,
                ref,
                args_to_string(job.args, job.kwargs),
                f" try={job.job_retries}" if job.job_retries > 1 else "",
            )
        )
        try:
            async with async_timeout.timeout(self.job_timeout):
                if task.bind:
                    result = await task.function(task, *(job.args or []), **(job.kwargs or {}))
                else:
                    result = await task.function(*(job.args or []), **(job.kwargs or {}))

            job_result.success = True
            job_result.finish_time = timezone.now()
            job.status = JobStatus.success
            logger.info("%6.2fs ← %s ● %s" % ((timestamp_ms_now() - start_ms) / 1000, ref, result))
            self.jobs_complete += 1

        except Exception as e:
            job_result.finish_time = timezone.now()
            self.jobs_failed += 1
            result = f"Run task error in NO.{job.job_retries} times, exc: {e}, retry after {self.job_retry_after} seconds"
            logger.error("%6.2fs ← %s ● %s" % ((timestamp_ms_now() - start_ms) / 1000, ref, result))

            if job.job_retries >= job.job_retry:
                t = (timestamp_ms_now() - to_ms_timestamp(job.enqueue_time)) / 1000
                logger.error("%6.2fs ! %s max retries %d exceeded" % (t, ref, job.job_retry))
                job.status = JobStatus.failed
            else:
                job.status = JobStatus.deferred
                job.job_retries = F("job_retries") + 1
                await redis.zadd(
                    DELAY_QUEUE, to_ms_timestamp(self.job_retry_after), f"{queue}:{job_id}"
                )
        finally:
            await self._xack(queue, msg_id)
            await job.save(update_fields=["status", "job_retries"])

        job_result.result = result
        await job_result.save()
        return job_result

    @property
    def worker_name(self):
        return f"{self.group_name}-{self.consumer_name}"

    async def _push_heartbeat(self, is_offline: bool = False):
        if is_offline:
            await self._redis.hdel(constants.WORKER_KEY, self.worker_name)
        else:
            value = {
                "queue": self.queue,
                "is_timer": isinstance(self, TimerWorker),
                "ms": timestamp_ms_now(),
            }
            await self._redis.hset(constants.WORKER_KEY, self.worker_name, value=json.dumps(value))

    async def _heartbeat(self):
        """
        keep alive in redis
        """
        while True:
            await self._push_heartbeat()
            await asyncio.sleep(constants.WORKER_HEARTBEAT_SECONDS)

    async def _pre_run(self):
        try:
            await self._redis.xgroup_create(
                self.queue, self.group_name, latest_id="$", mkstream=True
            )
        except BusyGroupError:
            pass
        if not self.consumer_name:
            async with await self._lock_manager.lock(constants.WORKER_KEY_LOCK):
                workers = await self._redis.hgetall(constants.WORKER_KEY)
                length = len(
                    list(
                        filter(
                            lambda item: not json.loads(item[1]).get("is_timer"), workers.items()
                        )
                    )
                )
                self.consumer_name = length
                await self._push_heartbeat()

    async def run(self):
        await self._pre_run()
        try:
            await asyncio.gather(self._main(), self._heartbeat())
        except asyncio.CancelledError:
            pass
        finally:
            await self._push_heartbeat(True)
            await self.close()

    async def close(self):
        await self.rearq.shutdown()
        await self.rearq.close()


class TimerWorker(Worker):
    def __init__(self, rearq: "ReArq"):
        super().__init__(rearq)
        self.consumer_name = "timer"
        self.queue = DELAY_QUEUE

    async def _run_at_start(self):
        jobs = []
        p = self._redis.pipeline()
        for function, task in CronTask.get_cron_tasks().items():
            if task.run_at_start:
                logger.info(f"{function}() <- run at start")
                job_id = uuid4().hex
                jobs.append(
                    Job(
                        task=function,
                        job_retry=self.job_retry,
                        queue=task.queue,
                        job_id=job_id,
                        enqueue_time=timezone.now(),
                        job_retry_after=self.job_retry_after,
                        status=JobStatus.queued,
                    )
                )
                p.xadd(task.queue, {"job_id": job_id})
                self.jobs_complete += 1
        if jobs:
            await Job.bulk_create(jobs)
            await p.execute()

    async def _main(self) -> None:
        tasks = list(CronTask.get_cron_tasks().keys())
        tasks.remove(check_pending_msgs.__name__)
        logger.info("Start timer success")
        logger.add(f"logs/worker-{self.consumer_name}.log", rotation="00:00")
        logger.info(f"Registered timer tasks: {', '.join(tasks)}")

        await self.log_redis_info()
        await self.rearq.startup()
        await self._run_at_start()

        async for _ in poll():
            await self._poll_iteration()
            await self.run_cron()

    async def _pre_run(self):
        async with await self._lock_manager.lock(constants.WORKER_KEY_LOCK):
            workers_info = await self._redis.hgetall(constants.WORKER_KEY)
            for worker_name, value in workers_info.items():
                value = json.loads(value)
                time = ms_to_datetime(value["ms"])
                is_offline = (
                    timezone.now() - time
                ).seconds > constants.WORKER_HEARTBEAT_SECONDS + 10
                if value.get("is_timer") and not is_offline:
                    msg = f"There is a timer worker `{worker_name}` already, you can only start one timer worker"

                    logger.error(msg)
                    raise UsageError(msg)
            else:
                await self._push_heartbeat()

    async def run_cron(self):
        """
        run cron task
        :return:
        """
        redis = self._redis
        cron_tasks = CronTask.get_cron_tasks()
        p = redis.pipeline()
        execute = False
        jobs = []
        for function, task in cron_tasks.items():
            if timestamp_ms_now() >= task.next_run:
                execute = True
                job_id = uuid4().hex
                if task.function == check_pending_msgs:
                    asyncio.ensure_future(
                        check_pending_msgs(task, task.queue, self.group_name, self.job_timeout)
                    )
                else:
                    logger.info(f"{task.function.__name__}()")
                    jobs.append(
                        Job(
                            task=function,
                            job_retry=self.job_retry,
                            queue=task.queue,
                            job_id=job_id,
                            enqueue_time=timezone.now(),
                            job_retry_after=self.job_retry_after,
                            status=JobStatus.queued,
                        )
                    )
                    p.xadd(task.queue, {"job_id": job_id})
                    self.jobs_complete += 1
                task.set_next()
        if jobs:
            await Job.bulk_create(jobs)
        if execute:
            await p.execute()

    async def _poll_iteration(self):
        """
        get delay task and put to queue
        :return:
        """
        redis = self._redis
        now = timestamp_ms_now()
        jobs_id = await redis.zrangebyscore(
            DELAY_QUEUE, offset=0, count=self.queue_read_limit, max=now
        )
        if not jobs_id:
            return
        p = redis.pipeline()
        for job_id in jobs_id:  # type:str
            separate = job_id.rindex(":")
            queue, job_id = job_id[:separate], job_id[separate + 1 :]  # noqa:
            p.xadd(queue, {"job_id": job_id})
        p.zrem(DELAY_QUEUE, *jobs_id)
        await p.execute()



import json
import os

import aiofiles
from fastapi import APIRouter, Depends
from starlette.requests import Request
from tortoise import timezone
from tortoise.functions import Count

from rearq import ReArq, constants
from rearq.server import templates
from rearq.server.depends import get_rearq
from rearq.server.models import JobResult
from rearq.utils import ms_to_datetime

router = APIRouter()


@router.get("", include_in_schema=False)
async def get_workers(request: Request, rearq: ReArq = Depends(get_rearq)):
    redis = rearq.redis
    workers_info = await redis.hgetall(constants.WORKER_KEY)
    workers = []
    for worker_name, value in workers_info.items():
        job_stat = (
            await JobResult.filter(worker=worker_name)
            .annotate(count=Count("job_id"))
            .group_by("job__status")
            .values("count", status="job__status",)
        )
        item = {"name": worker_name, "job_stat": {job["status"]: job["count"] for job in job_stat}}
        item.update(json.loads(value))
        time = ms_to_datetime(item["ms"])
        item["time"] = time
        item["is_offline"] = (
            timezone.now() - time
        ).seconds > constants.WORKER_HEARTBEAT_SECONDS + 10

        workers.append(item)
    return templates.TemplateResponse(
        "worker.html", {"request": request, "page_title": "worker", "workers": workers}
    )


@router.delete("")
async def delete_worker(name: str, rearq: ReArq = Depends(get_rearq)):
    redis = rearq.redis
    return await redis.hdel(constants.WORKER_KEY, name)


@router.get("/log")
async def logs(name: str):
    log_file = os.path.join(constants.BASE_DIR, "logs", f"worker-{name}.log")
    async with aiofiles.open(log_file, mode="r") as f:
        content = await f.read()
    return content



from tortoise import Model, fields

from rearq import job


class Job(Model):
    task = fields.CharField(max_length=200)
    args = fields.JSONField(null=True)
    kwargs = fields.JSONField(null=True)
    job_retry = fields.IntField()
    job_retries = fields.IntField(default=0)
    job_retry_after = fields.IntField()
    job_id = fields.CharField(max_length=200, unique=True)
    enqueue_time = fields.DatetimeField()
    expire_time = fields.DatetimeField(null=True)
    status = fields.CharEnumField(job.JobStatus)

    class Meta:
        ordering = ['-id']


class JobResult(Model):
    job: fields.ForeignKeyRelation[Job] = fields.ForeignKeyField("models.Job")
    worker = fields.CharField(max_length=200)
    success = fields.BooleanField(default=False)
    msg_id = fields.CharField(max_length=200, unique=True)
    result = fields.TextField(null=True)
    start_time = fields.DatetimeField(null=True)
    finish_time = fields.DatetimeField(null=True)

    class Meta:
        ordering = ['-id']



import os

QUEUE_KEY_PREFIX = "rearq:queue:"
DEFAULT_QUEUE: str = f"{QUEUE_KEY_PREFIX}default"
DELAY_QUEUE = f"{QUEUE_KEY_PREFIX}delay"
WORKER_KEY = "rearq:worker"
WORKER_KEY_LOCK = "rearq:worker:lock"
WORKER_HEARTBEAT_SECONDS = 10
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))



from fastapi import APIRouter, Depends
from starlette.requests import Request

from rearq import CronTask, ReArq
from rearq.server import templates
from rearq.server.depends import get_rearq
from rearq.server.models import JobResult
from rearq.utils import ms_to_datetime

router = APIRouter()


@router.get("", include_in_schema=False)
async def get_tasks(request: Request, rearq: ReArq = Depends(get_rearq)):
    task_map = rearq.task_map
    tasks = []
    cron_tasks = []
    for task_name, task in task_map.items():
        item = {
            "name": task_name,
            "queue": task.queue,
        }
        job_result = await JobResult.filter(job__task=task_name).order_by("-id").first()
        if job_result:
            item["last_time"] = job_result.finish_time
        else:
            item["last_time"] = None
        if isinstance(task, CronTask):
            item["cron"] = task.cron
            task.set_next()
            item["next_time"] = ms_to_datetime(task.next_run)
            cron_tasks.append(item)
        else:
            tasks.append(item)
    return templates.TemplateResponse(
        "task.html",
        {"request": request, "page_title": "task", "tasks": tasks, "cron_tasks": cron_tasks},
    )



import os
from datetime import date

from starlette.templating import Jinja2Templates

from rearq import constants, version

templates = Jinja2Templates(
    directory=os.path.join(constants.BASE_DIR, "rearq", "server", "templates")
)
templates.env.globals["REAEQ_VERSION"] = version.VERSION
templates.env.globals["NOW_YEAR"] = date.today().year



from fastapi import APIRouter, Depends
from pypika.functions import Date
from starlette.requests import Request
from tortoise.functions import Count, Function

from rearq import ReArq, constants
from rearq.server import templates
from rearq.server.depends import get_rearq
from rearq.server.models import Job, JobResult

router = APIRouter()


class ToDate(Function):
    database_func = Date


@router.get("/", include_in_schema=False)
async def index(request: Request, rearq: ReArq = Depends(get_rearq)):
    task_map = rearq.task_map
    task_num = len(task_map)
    workers_info = await rearq.redis.hgetall(constants.WORKER_KEY)
    worker_num = len(workers_info)
    run_times = await JobResult.all().count()
    result = (
        await Job.all()
        .annotate(count=Count("id"), date=ToDate("enqueue_time"))
        .group_by("date", "status")
        .order_by("date")
        .values("date", "status", "count")
    )
    x_axis = []
    series = [
        {
            "name": "Deferred",
            "type": "line",
            "data": [],
            "stack": "total",
            "label": {"show": "true"},
        },
        {
            "name": "Queued",
            "type": "line",
            "stack": "total",
            "data": [],
            "label": {"show": "true"},
        },
        {
            "name": "InProgress",
            "type": "line",
            "stack": "total",
            "data": [],
            "label": {"show": "true"},
        },
        {
            "name": "Success",
            "type": "line",
            "stack": "total",
            "data": [],
            "label": {"show": "true"},
        },
        {
            "name": "Failed",
            "type": "line",
            "stack": "total",
            "data": [],
            "label": {"show": "true"},
        },
        {
            "name": "Expired",
            "type": "line",
            "stack": "total",
            "data": [],
            "label": {"show": "true"},
        },
    ]
    for item in result:
        date = str(item.get("date"))
        if date not in x_axis:
            x_axis.append(date)
        count = item.get("count")
        status = item.get("status")
        if status == "deferred":
            series[0]["data"].append(count)
        elif status == "queued":
            series[1]["data"].append(count)
        elif status == "in_progress":
            series[2]["data"].append(count)
        elif status == "success":
            series[3]["data"].append(count)
        elif status == "failed":
            series[4]["data"].append(count)
        elif status == "expired":
            series[5]["data"].append(count)
    return templates.TemplateResponse(
        "dashboard.html",
        {
            "request": request,
            "page_title": "dashboard",
            "worker_num": worker_num,
            "task_num": task_num,
            "run_times": run_times,
            "x_axis": x_axis,
            "series": series,
        },
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rearq', 'rearq.server', 'rearq.server.routes']

package_data = \
{'': ['*'], 'rearq.server': ['static/*', 'templates/*', 'templates/widgets/*']}

install_requires = \
['aiofiles',
 'aioredis',
 'aioredlock',
 'click',
 'crontab',
 'fastapi',
 'jinja2',
 'loguru',
 'pydantic',
 'tortoise-orm',
 'uvicorn[standard]']

extras_require = \
{'mysql': ['asyncmy'], 'postgres': ['asyncpg']}

entry_points = \
{'console_scripts': ['rearq = rearq.cli:main']}

setup_kwargs = {
    'name': 'rearq',
    'version': '0.2.3',
    'description': 'Rewrite arq and make improvement.',
    'long_description': '# ReArq\n\n[![image](https://img.shields.io/pypi/v/rearq.svg?style=flat)](https://pypi.python.org/pypi/rearq)\n[![image](https://img.shields.io/github/license/long2ice/rearq)](https://github.com/long2ice/rearq)\n[![image](https://github.com/long2ice/rearq/workflows/pypi/badge.svg)](https://github.com/long2ice/rearq/actions?query=workflow:pypi)\n[![image](https://github.com/long2ice/rearq/workflows/ci/badge.svg)](https://github.com/long2ice/rearq/actions?query=workflow:ci)\n\n## Introduction\n\nReArq is a distributed task queue with asyncio and redis, which rewrite from [arq](https://github.com/samuelcolvin/arq)\nto make improvement and include web interface.\n\nYou can try [Demo Online](https://demo-rearq.long2ice.cn) here.\n\n## Screenshots\n\n![dashboard](./images/dashboard.png)\n![worker](./images/worker.png)\n![task](./images/task.png)\n![job](./images/job.png)\n![result](./images/result.png)\n\n## Install\n\nUse MySQL backend:\n\n```shell\n> pip install rearq[mysql]\n```\n\nUse PostgreSQL backend:\n\n```shell\n> pip install rearq[postgres]\n```\n\n## Quick Start\n\n### Task Definition\n\n```python\n# main.py\nfrom rearq import ReArq\nfrom rearq.server import models\n\nconfig = {\n    "connections": {\n        "default": f"mysql://root:123456@127.0.0.1:3306/rearq"\n    },\n    "apps": {"models": {"models": [models], "default_connection": "default"}},\n    "use_tz": True,\n}\n\nrearq = ReArq(tortoise_config=config)\n\n\n@rearq.on_shutdown\nasync def on_shutdown():\n    # you can do some clean work here like close db and so on...\n    print("shutdown")\n\n\n@rearq.on_startup\nasync def on_startup():\n    # you should do some initialization work here, such tortoise-orm init and so on...\n    print("startup")\n\n\n@rearq.task(queue="myqueue")\nasync def add(self, a, b):\n    return a + b\n\n\n@rearq.task(cron="*/5 * * * * * *")  # run task per 5 seconds\nasync def timer(self):\n    return "timer"\n```\n\n### Run rearq worker\n\n```shell\n> rearq main:rearq worker -q myqueue\n```\n\n```log\n2021-03-29 09:54:50.464 | INFO     | rearq.worker:_main:95 - Start worker success with queue: rearq:queue:default\n2021-03-29 09:54:50.465 | INFO     | rearq.worker:_main:96 - Registered tasks: add, sleep, timer_add\n2021-03-29 09:54:50.465 | INFO     | rearq.worker:log_redis_info:86 - redis_version=6.2.1 mem_usage=1.43M clients_connected=5 db_keys=6\n```\n\n### Run rearq timer\n\nIf you have timing task or delay task, you should run another command also:\n\n```shell\n> rearq main:rearq timer\n```\n\n```log\n2021-03-29 09:54:43.878 | INFO     | rearq.worker:_main:275 - Start timer success\n2021-03-29 09:54:43.887 | INFO     | rearq.worker:_main:277 - Registered timer tasks: timer_add\n2021-03-29 09:54:43.894 | INFO     | rearq.worker:log_redis_info:86 - redis_version=6.2.1 mem_usage=1.25M clients_connected=2 db_keys=6\n```\n\n### Integration in FastAPI\n\n```python\nfrom fastapi import FastAPI\n\napp = FastAPI()\n\n\n@app.on_event("startup")\nasync def startup() -> None:\n    await rearq.init()\n\n\n@app.on_event("shutdown")\nasync def shutdown() -> None:\n    await rearq.close()\n\n\n# then run task in view\n@app.get("/test")\nasync def test():\n    job = await add.delay(args=(1, 2))\n    return job.info()\n```\n\n## Start server\n\nYou can start the web interface.\n\n```shell\n> rearq main:rearq server\nUsage: rearq server [OPTIONS]\n\n  Start rest api server.\n\nOptions:\n  --host TEXT         Listen host.  [default: 0.0.0.0]\n  -p, --port INTEGER  Listen port.  [default: 8000]\n  -h, --help          Show this message and exit..\n```\n\nAfter server run, you can visit [https://127.0.0.1:8000/docs](https://127.0.0.1:8000/docs) to see all apis\nand [https://127.0.0.1:8000](https://127.0.0.1:8000) to see web interface.\n\n## ThanksTo\n\n- [arq](https://github.com/samuelcolvin/arq), Fast job queuing and RPC in python with asyncio and redis.\n\n## License\n\nThis project is licensed under the [MIT](https://github.com/long2ice/rearq/blob/master/LICENSE) License.\n',
    'author': 'long2ice',
    'author_email': 'long2ice@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/long2ice/rearq.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import os

from fastapi import FastAPI, HTTPException
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.staticfiles import StaticFiles

from .. import constants
from .routes import router

app = FastAPI(title="API docs of ReArq")
app.mount(
    "/static",
    StaticFiles(directory=os.path.join(constants.BASE_DIR, "rearq", "server", "static")),
    name="static",
)
app.include_router(router)


@app.exception_handler(HTTPException)
async def http_exception_handler(request: Request, exc: HTTPException):
    return JSONResponse(status_code=exc.status_code, content={"msg": exc.detail},)



from typing import Optional

from fastapi import APIRouter, Depends, HTTPException
from starlette.requests import Request
from starlette.status import HTTP_404_NOT_FOUND, HTTP_409_CONFLICT

from rearq import ReArq, constants
from rearq.job import JobStatus
from rearq.server import templates
from rearq.server.depends import get_pager, get_rearq
from rearq.server.models import Job, JobResult
from rearq.server.responses import JobListOut, JobOut
from rearq.server.schemas import AddJobIn, UpdateJobIn

router = APIRouter()


@router.get("", include_in_schema=False)
async def job_page(request: Request, rearq=Depends(get_rearq)):
    workers_info = await rearq.redis.hgetall(constants.WORKER_KEY)
    return templates.TemplateResponse(
        "job.html",
        {
            "request": request,
            "page_title": "job",
            "tasks": rearq.task_map.keys(),
            "workers": workers_info.keys(),
            "job_status": JobStatus,
        },
    )


@router.get("/data", response_model=JobListOut)
async def get_results(
    task: Optional[str] = None,
    job_id: Optional[str] = None,
    start_time: Optional[str] = None,
    end_time: Optional[str] = None,
    status: Optional[str] = None,
    pager=Depends(get_pager),
):
    qs = Job.all()
    if task:
        qs = qs.filter(task=task)
    if job_id:
        qs = qs.filter(job_id=job_id)
    if start_time:
        qs = qs.filter(enqueue_time__gte=start_time)
    if end_time:
        qs = qs.filter(enqueue_time__lte=end_time)
    if status:
        qs = qs.filter(status=status)
    results = await qs.limit(pager[0]).offset(pager[1])
    return {"rows": results, "total": await qs.count()}


@router.get("/result")
async def get_job_result(job_id: str):
    return await JobResult.get_or_none(job_id=job_id)


@router.put("")
async def update_job(update_job_in: UpdateJobIn):
    job = await Job.get_or_none(job_id=update_job_in.job_id)
    if not job:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail="Can't find job")
    if job.status in [JobStatus.queued, JobStatus.deferred]:
        await job.update_from_dict(update_job_in.dict(exclude_unset=True)).save()
    else:
        raise HTTPException(status_code=HTTP_409_CONFLICT, detail="Can't update job")


@router.delete("")
async def delete_job(ids: str):
    return await Job.filter(id__in=ids.split(",")).delete()


@router.post("", response_model=JobOut)
async def add_job(add_job_in: AddJobIn, rearq: ReArq = Depends(get_rearq)):
    task = rearq.task_map.get(add_job_in.task)
    if not task:
        raise HTTPException(status_code=HTTP_404_NOT_FOUND, detail="No such task")
    job = await task.delay(**add_job_in.dict(exclude={"task"}))
    return job



from typing import Optional

from fastapi import APIRouter, Depends
from starlette.requests import Request

from rearq import ReArq, constants
from rearq.server import templates
from rearq.server.depends import get_pager, get_rearq
from rearq.server.models import JobResult
from rearq.server.responses import JobResultListOut

router = APIRouter()


@router.get("/data", response_model=JobResultListOut)
async def get_results(
    task: Optional[str] = None,
    job_id: Optional[str] = None,
    start_time: Optional[str] = None,
    end_time: Optional[str] = None,
    worker: Optional[str] = None,
    success: Optional[str] = None,
    pager=Depends(get_pager),
):
    qs = JobResult.all().select_related("job")
    if task:
        qs = qs.filter(job__task=task)
    if job_id:
        qs = qs.filter(job__job_id=job_id)
    if start_time:
        qs = qs.filter(start_time__gte=start_time)
    if end_time:
        qs = qs.filter(start_time__lte=end_time)
    if worker:
        qs = qs.filter(worker=worker)
    if success:
        qs = qs.filter(success=success == "1")
    results = await qs.limit(pager[0]).offset(pager[1])
    return {"rows": results, "total": await qs.count()}


@router.delete("")
async def delete_result(ids: str):
    return await JobResult.filter(id__in=ids.split(",")).delete()


@router.get("", include_in_schema=False)
async def result(
    request: Request, rearq: ReArq = Depends(get_rearq),
):
    workers_info = await rearq.redis.hgetall(constants.WORKER_KEY)

    return templates.TemplateResponse(
        "result.html",
        {
            "request": request,
            "page_title": "result",
            "tasks": rearq.task_map.keys(),
            "workers": workers_info.keys(),
        },
    )
